from dataclasses import dataclass
from typing import Optional


@dataclass
class SourceDiff:
    """An object describing a single changed file.

    - `path` -- the path to the file;
    - `old_path` -- the previous path to the file; differs from `path` if the
      file was renamed;
    - `diff` -- difference between versions (if present).
    """

    path: str
    old_path: str = ""
    diff: Optional[str] = None

    def __post_init__(self):
        if not self.old_path:
            self.old_path = self.path



from typing import Tuple, Sequence, List

from sqlfluff.api.simple import get_simple_config
from sqlfluff.core import Linter
from sqlfluff.dialects.dialect_ansi import StatementSegment

from migration_lint.sql.constants import StatementType
from migration_lint.sql.operations import find_matching_segment
from migration_lint.sql.rules import (
    BACKWARD_INCOMPATIBLE_OPERATIONS,
    BACKWARD_COMPATIBLE_OPERATIONS,
    DATA_MIGRATION_OPERATIONS,
    RESTRICTED_OPERATIONS,
    IGNORED_OPERATIONS,
)


def classify_migration(raw_sql: str) -> Sequence[Tuple[str, StatementType]]:
    """Classify migration statements."""

    linter = Linter(config=get_simple_config(dialect="postgres"))
    result = linter.parse_string(raw_sql)
    parsed = result.root_variant()
    if not parsed or not parsed.tree:
        raise RuntimeError(f"Can't parse SQL from string: {raw_sql}")

    unparsable_parts = [part for part in parsed.tree.recursive_crawl("unparsable")]
    if not parsed or unparsable_parts:
        errors = [
            str(e.description)
            for e in parsed.lexing_violations + parsed.parsing_violations
        ]
        raise RuntimeError(
            f"Can't parse SQL from string: {raw_sql}\n"
            f"Errors: \n" + "\n- ".join(errors)
        )

    statements = []
    statements_types = []
    for statement in parsed.tree.recursive_crawl("statement"):
        statements.append(statement)

    for statement in statements:
        statement_type = classify_statement(statement, context=statements)  # type: ignore
        if statement_type == StatementType.IGNORED:
            continue
        statements_types.append((statement.raw_normalized(), statement_type))

    return statements_types


def classify_statement(
    statement: StatementSegment, context: List[StatementSegment]
) -> StatementType:
    """
    Classify an SQL statement using predefined locators.
    :param statement: statement to classify
    :param context: all statements in the same migration
    :return:
    """

    statement_operations_map = {
        StatementType.IGNORED: IGNORED_OPERATIONS,
        StatementType.DATA_MIGRATION: DATA_MIGRATION_OPERATIONS,
        StatementType.BACKWARD_COMPATIBLE: BACKWARD_COMPATIBLE_OPERATIONS,
        StatementType.BACKWARD_INCOMPATIBLE: BACKWARD_INCOMPATIBLE_OPERATIONS,
        StatementType.RESTRICTED: RESTRICTED_OPERATIONS,
    }

    for statement_type, operations_locators in statement_operations_map.items():
        for locator in operations_locators:
            found_segment = find_matching_segment(
                segment=statement,
                locator=locator,
                context=context,
            )
            if found_segment:
                return statement_type

    return StatementType.UNSUPPORTED



import json
import os
from typing import Any, Sequence
from urllib.parse import urljoin
from urllib.request import Request, urlopen

from migration_lint import logger
from migration_lint.source_loader.base import BaseSourceLoader
from migration_lint.source_loader.model import SourceDiff


class GitlabBranchLoader(BaseSourceLoader):
    """A loader to obtain files changed in the given branch comparing to
    master.
    """

    NAME = "gitlab_branch"

    def __init__(
        self,
        branch: str,
        project_id: str,
        gitlab_api_key: str,
        gitlab_instance: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.base_url = gitlab_instance
        self.branch = branch
        self.project_id = project_id
        self.gitlab_api_key = gitlab_api_key
        self.default_branch = os.environ.get("CI_DEFAULT_BRANCH", "master")

        if not self.branch or not self.project_id or not self.gitlab_api_key:
            raise RuntimeError(
                f"You must specify branch, project_id and gitlab_api_key "
                f"to use GitlabBranchLoader (branch={self.branch}, project_id={self.project_id}, "
                f"gitlab_api_key={self.gitlab_api_key})"
            )

    def get_changed_files(self) -> Sequence[SourceDiff]:
        """Return a list of changed files."""

        logger.info(
            f"### Getting changed files from {self.default_branch} <-> {self.branch}"
        )

        endpoint = f"projects/{self.project_id}/repository/compare?"
        query_params = f"from={self.default_branch}&to={self.branch}"
        url = urljoin(f"{self.base_url}/api/v4/", endpoint + query_params)
        req = Request(url, headers={"PRIVATE-TOKEN": self.gitlab_api_key})
        with urlopen(req) as resp:
            diffs = json.loads(resp.read().decode("utf-8"))["diffs"]

        return [
            SourceDiff(
                diff=diff["diff"],
                path=diff["new_path"],
                old_path=diff["old_path"],
            )
            for diff in diffs
            if not diff["deleted_file"]
            and (not self.only_new_files or self.only_new_files and diff["new_file"])
        ]


class GitlabMRLoader(BaseSourceLoader):
    """A loader to obtain files changed in the given MR."""

    NAME = "gitlab_mr"

    def __init__(
        self,
        mr_id: str,
        project_id: str,
        gitlab_api_key: str,
        gitlab_instance: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.base_url = gitlab_instance
        self.mr_id = mr_id
        self.project_id = project_id
        self.gitlab_api_key = gitlab_api_key

        if not self.mr_id or not self.project_id or not self.gitlab_api_key:
            raise RuntimeError(
                f"You must specify mr_id, project_id and gitlab_api_key "
                f"to use GitlabMRLoader (mr_id={self.mr_id}, project_id={self.project_id}, "
                f"gitlab_api_key={self.gitlab_api_key})"
            )

    def get_changed_files(self) -> Sequence[SourceDiff]:
        """Return a list of changed files."""

        logger.info(f"### Getting changed files from MR: {self.mr_id}")

        endpoint = f"projects/{self.project_id}/merge_requests/{self.mr_id}"
        url = urljoin(f"{self.base_url}/api/v4/", endpoint)
        req = Request(url, headers={"PRIVATE-TOKEN": self.gitlab_api_key})
        with urlopen(req) as resp:
            mr_info = json.loads(resp.read().decode("utf-8"))

        logger.info(f"MR link: {mr_info['web_url']}")

        endpoint = f"projects/{self.project_id}/merge_requests/{self.mr_id}/diffs"
        url = urljoin(f"{self.base_url}/api/v4/", endpoint)
        req = Request(url, headers={"PRIVATE-TOKEN": self.gitlab_api_key})
        with urlopen(req) as resp:
            diffs = json.loads(resp.read().decode("utf-8"))

        return [
            SourceDiff(
                diff=diff["diff"],
                path=diff["new_path"],
                old_path=diff["old_path"],
            )
            for diff in diffs
            if not diff["deleted_file"]
            and (not self.only_new_files or self.only_new_files and diff["new_file"])
        ]



import os.path

from migration_lint import logger
from migration_lint.extractor.django import DjangoExtractor

import django.apps
from django.core.management import call_command, CommandError


class DjangoManagementExtractor(DjangoExtractor):
    """Migrations extractor for Django migrations for management command."""

    NAME = "django_management"

    def extract_sql(self, migration_path: str) -> str:
        """Extract raw SQL from the migration file."""

        parts = migration_path.split("/")
        file_name = parts[-1]
        app = parts[-3]
        migration_name = file_name.replace(".py", "")

        # handle subapp app name
        for app_config in django.apps.apps.get_app_configs():
            app_path = os.path.relpath(app_config.path, ".")
            app_migrations_path = os.path.join(app_path, "migrations")
            if migration_path.startswith(app_migrations_path):
                app = app_config.label
                break

        logger.info(
            f"Extracting sql for migration: app={app}, migration_name={migration_name}"
        )

        try:
            return call_command("sqlmigrate", app, migration_name)
        except CommandError:
            logger.error(
                f"Failed to extract SQL for migration app={app}, migration_name={migration_name}"
            )
            return ""



import enum

from migration_lint.util.colors import grey, green, red, yellow


class StatementType(str, enum.Enum):
    """Types of migration statements."""

    IGNORED = "ignored"
    BACKWARD_COMPATIBLE = "backward_compatible"
    BACKWARD_INCOMPATIBLE = "backward_incompatible"
    DATA_MIGRATION = "data_migration"
    RESTRICTED = "restricted"
    UNSUPPORTED = "unsupported"

    @property
    def colorized(self):
        if self is StatementType.IGNORED:
            return grey(self)
        elif self is StatementType.BACKWARD_COMPATIBLE:
            return green(self)
        elif self is StatementType.BACKWARD_INCOMPATIBLE:
            return red(self)
        elif self is StatementType.DATA_MIGRATION:
            return red(self)
        elif self is StatementType.RESTRICTED:
            return yellow(self)
        elif self is StatementType.UNSUPPORTED:
            return red(self)



import os

import click

from migration_lint import logger
from migration_lint.analyzer import Analyzer, CompatibilityLinter, SquawkLinter
from migration_lint.extractor import Extractor, DjangoExtractor
from migration_lint.source_loader import SourceLoader, LocalLoader
from migration_lint.util.env import get_bool_env


@click.command()
# Base setup
@click.option(
    "--loader",
    "loader_type",
    help="loader type (where to take source files changes)",
    type=click.Choice(SourceLoader.names(), case_sensitive=False),
    default=os.getenv("LOADER_TYPE", LocalLoader.NAME),
)
@click.option(
    "--extractor",
    "extractor_type",
    help="extractor type (how to extract SQL from migrations)",
    type=click.Choice(Extractor.names(), case_sensitive=False),
    default=os.getenv("EXTRACTOR", DjangoExtractor.NAME),
)
@click.option(
    "--only-new-files",
    help="lint only new files, ignore changes in existing files",
    default=get_bool_env("ONLY_NEW_FILES", True),
)
# gitlab-specific arguments
@click.option(
    "--project-id",
    help="GitLab project id (repo)",
    default=os.getenv("CI_PROJECT_ID"),
)
@click.option(
    "--gitlab-instance",
    help="GitLab instance instance (protocol://host:port)",
    default=os.getenv("CI_SERVER_URL"),
)
@click.option(
    "--gitlab-api-key",
    help="api key for GitLab API",
    default=os.getenv("CI_DEPLOY_GITLAB_TOKEN"),
)
@click.option(
    "--branch",
    help="branch to compare",
    default=os.getenv(
        "CI_MERGE_REQUEST_SOURCE_BRANCH_NAME", os.getenv("CI_COMMIT_BRANCH")
    ),
)
@click.option(
    "--mr-id",
    help="integer merge request id",
    default=os.getenv("CI_MERGE_REQUEST_ID"),
)
@click.option(
    "--squawk-config-path",
    "squawk_config_path",
    help="squawk config path",
    default=os.getenv("MIGRATION_LINTER_SQUAWK_CONFIG_PATH"),
)
@click.option(
    "--squawk-pg-version",
    "squawk_pg_version",
    help="squawk version of PostgreSQL",
    default=os.getenv("MIGRATION_LINTER_SQUAWK_PG_VERSION"),
)
@click.option(
    "--alembic-command",
    "alembic_command",
    help="command to get Alembic migrations sql",
    default=os.getenv("MIGRATION_LINTER_ALEMBIC_COMMAND"),
)
@click.option(
    "--ignore-extractor-fail",
    "ignore_extractor_fail",
    is_flag=True,
    help="Don't fail the whole linter if extraction of sql fails",
    default=os.getenv("MIGRATION_LINTER_IGNORE_EXTRACTOR_FAIL", False),
)
@click.option(
    "--ignore-extractor-not-found",
    "ignore_extractor_not_found",
    is_flag=True,
    help="Don't fail the whole linter if extraction went fine, but info about particular migration couldn't be found",
    default=os.getenv("MIGRATION_LINTER_IGNORE_EXTRACTOR_NOT_FOUND", False),
)
def main(
    loader_type: str,
    extractor_type: str,
    squawk_config_path: str,
    squawk_pg_version: str,
    **kwargs,
) -> None:
    logger.info("Start analysis..")

    loader = SourceLoader.get(loader_type)(**kwargs)
    extractor = Extractor.get(extractor_type)(**kwargs)
    analyzer = Analyzer(
        loader=loader,
        extractor=extractor,
        linters=[
            CompatibilityLinter(),
            SquawkLinter(
                config_path=squawk_config_path,
                pg_version=squawk_pg_version,
            ),
        ],
    )
    analyzer.analyze()


if __name__ == "__main__":
    main()



from typing import Optional, List

from sqlfluff.core.parser import BaseSegment, PositionMarker
from sqlfluff.dialects.dialect_ansi import StatementSegment

from migration_lint.sql.model import SegmentLocator


def find_matching_segment(
    segment: BaseSegment,
    locator: SegmentLocator,
    min_position: Optional[PositionMarker] = None,
    context: Optional[List[StatementSegment]] = None,
) -> Optional[BaseSegment]:
    """Find matching segment by the given locator starting with the given
    position.
    """

    for found in segment.recursive_crawl(locator.type):
        if (
            not locator.ignore_order
            and min_position
            and found.pos_marker
            and found.pos_marker < min_position
        ):
            continue
        attrs_match = True
        if locator.raw and locator.raw.upper() != found.raw.upper():
            attrs_match = False

        children_match = True
        if locator.children:
            min_position = None
            for child_locator in locator.children:
                child_found = find_matching_segment(
                    segment=found,
                    locator=child_locator,
                    min_position=min_position,
                )
                if (
                    not child_locator.inverted
                    and not child_found
                    or child_locator.inverted
                    and child_found
                ):
                    children_match = False
                if child_found:
                    min_position = child_found.pos_marker

        # unsafe segment still can be safe if paired with specific
        only_with_match = locator.only_with is None
        if locator.only_with and context:
            match_by_original = find_matching_segment(
                segment, locator.only_with.match_by
            )
            for context_statement in context:
                # searching in the same migration
                found_context_segment = find_matching_segment(
                    context_statement, locator.only_with.locator
                )
                if not found_context_segment:
                    continue

                # matching in context
                # (for example checking table that being created
                # is the same as being altered)
                match_by_context = find_matching_segment(
                    found_context_segment, locator.only_with.match_by
                )
                if (
                    match_by_original
                    and match_by_context
                    and match_by_original.raw_normalized()
                    == match_by_context.raw_normalized()
                ):
                    only_with_match = True

        if attrs_match and children_match and only_with_match:
            return found

    return None



from enum import Enum


class ColorCode(Enum):
    blue = "\x1b[1;34m"
    grey = "\x1b[38;21m"
    green = "\x1b[1;32m"
    red = "\x1b[31;21m"
    yellow = "\x1b[33;21m"
    reset = "\x1b[0m"


def colorize(color: ColorCode, msg: str) -> str:
    return f"{color.value}{msg}{ColorCode.reset.value}"


def grey(msg: str) -> str:
    return colorize(ColorCode.grey, msg)


def green(msg: str) -> str:
    return colorize(ColorCode.green, msg)


def yellow(msg: str) -> str:
    return colorize(ColorCode.yellow, msg)


def red(msg: str) -> str:
    return colorize(ColorCode.red, msg)


def blue(msg: str) -> str:
    return colorize(ColorCode.blue, msg)



from migration_lint.extractor.raw_sql import RawSqlExtractor


class FlywayExtractor(RawSqlExtractor):
    """Migrations extractor for Flyway migrations."""

    NAME = "flyway"

    def is_migration(self, path: str) -> bool:
        """Check if the specified file is a migration."""

        return "/db/migration/" in path and path.endswith(".sql")

    def is_allowed_with_backward_incompatible_migration(self, path: str) -> bool:
        """Check if the specified file changes are allowed with
        backward-incompatible migrations.
        """

        return self.is_migration(path)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from migration_lint.source_loader.model import SourceDiff


@dataclass
class Migration:
    """A database migration file representation.

    - path -- path to the migration file;
    - raw_sql -- raw SQL representation of the migration.
    """

    path: str
    raw_sql: str


@dataclass
class ExtendedSourceDiff(SourceDiff):
    """An object desribing a single changed file.

    - `path` -- the path to the file;
    - `old_path` -- the previous path to the file; differs from `path` if the
      file was renamed;
    - `diff` -- difference between versions (if present).
    - `allowed_with_backward_incompatible` -- is this file is allowed with
      backward-incompatible migrations.
    """

    allowed_with_backward_incompatible: bool = False

    @classmethod
    def of_source_diff(
        cls,
        source_diff: SourceDiff,
        allowed_with_backward_incompatible,
    ) -> ExtendedSourceDiff:
        """Create an instance by the given source diff."""

        return cls(
            path=source_diff.path,
            old_path=source_diff.old_path,
            diff=source_diff.diff,
            allowed_with_backward_incompatible=allowed_with_backward_incompatible,
        )


@dataclass
class MigrationsMetadata:
    """Migrations metadata.

    - `changed_files` -- a list of changed files;
    - `migrations` -- a list of migrations.
    """

    changed_files: List[ExtendedSourceDiff] = field(default_factory=list)
    migrations: List[Migration] = field(default_factory=list)



from __future__ import annotations

import dataclasses
from typing import Optional, List


@dataclasses.dataclass
class SegmentLocator:
    """A locator object used to find segments.

    - `type` -- string used to match the type of operation (see sqlfluff types);
    - `raw` -- string used to match segment by raw SQL;
    - `children` -- list of locators used to match descendant segments (on any
      level);
    - `inverted` --- flag to perform an "inverted" match; works only for
      children, so it's actually "segment that doesn't contain specific
      descendant".
    - `ignore_order` --- by default linter looks for keywords in order,
      use this flag to search regardless
    - `only_with` --- if statement comes with this statement in one migration, it's safe,
      for example if we create table and add foreign key in one migration, it's safe,
      if we add foreign key on existing big table - it's not.
    """

    type: str
    raw: Optional[str] = None
    children: Optional[List[SegmentLocator]] = None
    inverted: bool = False
    ignore_order: bool = False
    only_with: Optional[ConditionalMatch] = None


@dataclasses.dataclass
class KeywordLocator(SegmentLocator):
    """A locator object used to find segments by keyword."""

    type: str = "keyword"


@dataclasses.dataclass
class ConditionalMatch:
    """
    An object to segment by condition
    (for example CREATE TABLE and ALTER TABLE for the same table).
    See sql/rules.py for examples.

    - `locator` -- what segment to look for.
    - `match_by` -- how to match two segments (by table name, column name etc);
    """

    locator: SegmentLocator
    match_by: SegmentLocator



from typing import List

from migration_lint.sql.model import ConditionalMatch, KeywordLocator, SegmentLocator

BACKWARD_INCOMPATIBLE_OPERATIONS: List[SegmentLocator] = [
    SegmentLocator(type="drop_sequence_statement"),
    # TODO: drop constraints first?
    SegmentLocator(type="drop_table_statement"),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="DROP"),
            KeywordLocator(raw="DEFAULT"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="SET"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="NULL"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="PRIMARY"),
            KeywordLocator(raw="KEY"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="CONSTRAINT"),
            KeywordLocator(raw="PRIMARY"),
            KeywordLocator(raw="KEY"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="DROP"),
            KeywordLocator(raw="COLUMN"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="RENAME"),
            KeywordLocator(raw="COLUMN"),
        ],
    ),
    SegmentLocator(type="truncate_table"),
    SegmentLocator(type="drop_type_statement"),
]

BACKWARD_COMPATIBLE_OPERATIONS: List[SegmentLocator] = [
    SegmentLocator(
        type="create_index_statement", children=[KeywordLocator(raw="CONCURRENTLY")]
    ),
    SegmentLocator(
        type="drop_index_statement", children=[KeywordLocator(raw="CONCURRENTLY")]
    ),
    SegmentLocator(
        type="alter_index_statement", children=[KeywordLocator(raw="RENAME")]
    ),
    SegmentLocator(
        type="reindex_statement_segment", children=[KeywordLocator(raw="CONCURRENTLY")]
    ),
    SegmentLocator(type="create_sequence_statement"),
    SegmentLocator(type="alter_sequence_statement"),
    SegmentLocator(type="create_table_statement"),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="NOT", inverted=True),
            KeywordLocator(raw="NULL"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="DROP"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="NULL"),
        ],
    ),
    # it's ok to do "SET NOT NULL" only after explicit "VALIDATE CONSTRAINT"
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="SET"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="NULL"),
        ],
        only_with=ConditionalMatch(
            SegmentLocator(
                type="alter_table_statement",
                children=[
                    KeywordLocator(raw="VALIDATE"),
                    KeywordLocator(raw="CONSTRAINT"),
                ],
            ),
            match_by=SegmentLocator(type="table_reference"),
        ),
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="NULL"),
            KeywordLocator(raw="DEFAULT", ignore_order=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="SET"),
            KeywordLocator(raw="DEFAULT"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="CONSTRAINT"),
            KeywordLocator(raw="UNIQUE"),
            KeywordLocator(raw="USING"),
            KeywordLocator(raw="INDEX"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="CONSTRAINT"),
        ],
        only_with=ConditionalMatch(
            locator=SegmentLocator(type="create_table_statement"),
            match_by=SegmentLocator(type="table_reference"),
        ),
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="DROP"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="NULL"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="DROP"),
            KeywordLocator(raw="CONSTRAINT"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="CONSTRAINT"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="VALID"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="VALIDATE"),
            KeywordLocator(raw="CONSTRAINT"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="FOREIGN"),
            KeywordLocator(raw="KEY"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="VALID"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="FOREIGN"),
            KeywordLocator(raw="KEY"),
        ],
        only_with=ConditionalMatch(
            locator=SegmentLocator(type="create_table_statement"),
            match_by=SegmentLocator(type="table_reference"),
        ),
    ),
    # change type to text is always safe
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="TYPE"),
            SegmentLocator(type="data_type", raw="TEXT"),
        ],
    ),
    # basic ADD COLUMN is default to NULL, so it's safe
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="NOT", inverted=True),
            KeywordLocator(raw="NULL", inverted=True),
            KeywordLocator(raw="PRIMARY", inverted=True),
            KeywordLocator(raw="KEY", inverted=True),
            KeywordLocator(raw="IDENTITY", inverted=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="GENERATED"),
            KeywordLocator(raw="IDENTITY"),
        ],
    ),
    SegmentLocator(type="create_statistics_statement"),
    SegmentLocator(type="analyze_statement"),
    SegmentLocator(type="reset_statement"),
    SegmentLocator(type="create_type_statement"),
    SegmentLocator(
        type="alter_type_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="VALUE"),
        ],
    ),
    SegmentLocator(
        type="insert_statement",
        children=[SegmentLocator(type="table_reference", raw="alembic_version")],
    ),
    SegmentLocator(
        type="update_statement",
        children=[SegmentLocator(type="table_reference", raw="alembic_version")],
    ),
    SegmentLocator(type="create_function_statement"),
    SegmentLocator(type="drop_function_statement"),
    SegmentLocator(type="create_trigger"),
    SegmentLocator(type="drop_trigger"),
]

RESTRICTED_OPERATIONS: List[SegmentLocator] = [
    SegmentLocator(
        type="create_index_statement",
        children=[KeywordLocator(raw="CONCURRENTLY", inverted=True)],
    ),
    SegmentLocator(
        type="drop_index_statement",
        children=[KeywordLocator(raw="CONCURRENTLY", inverted=True)],
    ),
    SegmentLocator(
        type="reindex_statement_segment",
        children=[KeywordLocator(raw="CONCURRENTLY", inverted=True)],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="CONSTRAINT"),
            KeywordLocator(raw="UNIQUE"),
            KeywordLocator(raw="USING", inverted=True),
            KeywordLocator(raw="INDEX", inverted=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="RENAME"),
            KeywordLocator(raw="COLUMN", inverted=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ALTER"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="TYPE"),
            SegmentLocator(type="data_type", raw="TEXT", inverted=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="CONSTRAINT"),
            KeywordLocator(raw="NOT", inverted=True),
            KeywordLocator(raw="VALID", inverted=True),
            KeywordLocator(raw="USING", inverted=True),
            KeywordLocator(raw="INDEX", inverted=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="FOREIGN"),
            KeywordLocator(raw="KEY"),
            KeywordLocator(raw="NOT", inverted=True),
            KeywordLocator(raw="VALID", inverted=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="NOT"),
            KeywordLocator(raw="NULL"),
            KeywordLocator(raw="DEFAULT", inverted=True, ignore_order=True),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="PRIMARY"),
            KeywordLocator(raw="KEY"),
        ],
    ),
    SegmentLocator(
        type="alter_table_statement",
        children=[
            KeywordLocator(raw="ADD"),
            KeywordLocator(raw="COLUMN"),
            KeywordLocator(raw="GENERATED"),
            KeywordLocator(raw="IDENTITY"),
        ],
    ),
]

DATA_MIGRATION_OPERATIONS = [
    SegmentLocator(type="update_statement"),
    SegmentLocator(type="insert_statement"),
    SegmentLocator(type="delete_statement"),
]

IGNORED_OPERATIONS = [
    SegmentLocator(type="select_statement"),
    SegmentLocator(type="set_statement"),
    SegmentLocator(
        type="transaction_statement", children=[KeywordLocator(raw="BEGIN")]
    ),
    SegmentLocator(type="transaction_statement", children=[KeywordLocator(raw="END")]),
    SegmentLocator(
        type="transaction_statement", children=[KeywordLocator(raw="COMMIT")]
    ),
]



from migration_lint.analyzer.base import Analyzer
from migration_lint.analyzer.compat import CompatibilityLinter
from migration_lint.analyzer.squawk import SquawkLinter

__all__ = (
    "Analyzer",
    "CompatibilityLinter",
    "SquawkLinter",
)



from __future__ import annotations

import abc
from typing import Sequence, List

from migration_lint import logger
from migration_lint.extractor.base import BaseExtractor
from migration_lint.extractor.model import ExtendedSourceDiff
from migration_lint.source_loader.base import BaseSourceLoader
from migration_lint.util.colors import green, red, yellow


# A migration ignore mark.
MANUALLY_IGNORE_ANNOTATION = "-- migration-lint: ignore"

CLASSIFICATION_LINK = "https://pandadoc.github.io/migration-lint/classification/"
IGNORE_LINK = "https://pandadoc.github.io/migration-lint/rules/#ignoring-statements"


class BaseLinter:
    """Base class for migration linters."""

    @abc.abstractmethod
    def lint(
        self,
        migration_sql: str,
        changed_files: List[ExtendedSourceDiff],
    ) -> List[str]:
        """Perform SQL migartion linting."""

        raise NotImplementedError()


class Analyzer:
    """Migrations analyzer."""

    def __init__(
        self,
        loader: BaseSourceLoader,
        extractor: BaseExtractor,
        linters: Sequence[BaseLinter],
    ) -> None:
        self.loader = loader
        self.extractor = extractor
        self.linters = linters

    def analyze(self) -> None:
        """Analyze migrations in files changed according to analyzer's source
        loader.
        """

        changed_files = self.loader.get_changed_files()
        metadata = self.extractor.create_metadata(changed_files)

        if not metadata.migrations:
            logger.info("Looks like you don't have any migration in MR.")
            return

        logger.info("")

        errors = []
        for migration in metadata.migrations:
            logger.info(green(f"Analyzing migration: {migration.path}\n"))

            if MANUALLY_IGNORE_ANNOTATION in migration.raw_sql:
                logger.info(yellow("Migration is ignored."))
                continue

            for linter in self.linters:
                errors.extend(
                    linter.lint(migration.raw_sql, metadata.changed_files),
                )

        logger.info("")

        if errors:
            logger.info(red("Errors found in migrations:\n"))
            for error in errors:
                logger.error(error)
            logger.info("")
            logger.info(
                f"See classification of statements if you need to fix: {CLASSIFICATION_LINK}"
            )
            logger.info(
                f"See how to ignore the linter for one migration: {IGNORE_LINK}"
            )
            exit(1)
        else:
            logger.info(green("Everything seems good!"))



import os


def get_bool_env(var_name, default: bool) -> bool:
    str_value = os.getenv(var_name, str(default)).lower()
    if str_value in ("true", "1", "yes"):
        return True
    elif str_value in ("false", "0", "no"):
        return False
    raise RuntimeError(f"Invalid boolean value for env {var_name}={str_value}")



from migration_lint.extractor.base import Extractor
from migration_lint.extractor.alembic import AlembicExtractor
from migration_lint.extractor.django import DjangoExtractor
from migration_lint.extractor.flyway import FlywayExtractor

__all__ = (
    "Extractor",
    "AlembicExtractor",
    "DjangoExtractor",
    "FlywayExtractor",
)



import os

from migration_lint import logger
from migration_lint.analyzer import Analyzer, CompatibilityLinter, SquawkLinter
from migration_lint.extractor import Extractor
from migration_lint.source_loader import SourceLoader, LocalLoader
from migration_lint.django.extractor.django_management import DjangoManagementExtractor  # noqa

from django.core.management.base import BaseCommand

from migration_lint.util.env import get_bool_env


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--loader",
            dest="loader_type",
            type=str,
            choices=SourceLoader.names(),
            default=os.getenv("LOADER_TYPE", LocalLoader.NAME),
            help="loader type (where to take source files changes)",
        )
        parser.add_argument(
            "--only-new-files",
            dest="only_new_files",
            action="store_true",
            default=get_bool_env("ONLY_NEW_FILES", True),
            help="lint only new files, ignore changes in existing files",
        )
        parser.add_argument(
            "--gitlab-instance",
            dest="gitlab_instance",
            type=str,
            default=os.getenv("CI_SERVER_URL"),
            help="GitLab instance instance (protocol://host:port)",
        )
        parser.add_argument(
            "--project-id",
            dest="project_id",
            type=str,
            default=os.getenv("CI_PROJECT_ID"),
            help="GitLab project id (repo)",
        )
        parser.add_argument(
            "--gitlab-api-key",
            dest="gitlab_api_key",
            type=str,
            default=os.getenv("CI_DEPLOY_GITLAB_TOKEN"),
            help="api key for GitLab API",
        )
        parser.add_argument(
            "--branch",
            dest="branch",
            type=str,
            default=os.getenv(
                "CI_MERGE_REQUEST_SOURCE_BRANCH_NAME", os.getenv("CI_COMMIT_BRANCH")
            ),
            help="branch to compare",
        )
        parser.add_argument(
            "--mr-id",
            dest="mr_id",
            type=str,
            default=os.getenv("CI_MERGE_REQUEST_ID"),
            help="integer merge request id",
        )
        parser.add_argument(
            "--squawk-config-path",
            dest="squawk_config_path",
            type=str,
            default=os.getenv("MIGRATION_LINTER_SQUAWK_CONFIG_PATH"),
            help="squawk config path",
        )
        parser.add_argument(
            "--squawk-pg-version",
            dest="squawk_pg_version",
            type=str,
            default=os.getenv("MIGRATION_LINTER_SQUAWK_PG_VERSION"),
            help="squawk version of PostgreSQL",
        )
        parser.add_argument(
            "--ignore-extractor-fail",
            dest="ignore_extractor_fail",
            action="store_true",
            default=get_bool_env("MIGRATION_LINTER_IGNORE_EXTRACTOR_FAIL", False),
            help="Don't fail the whole linter if extraction of sql fails",
        )
        parser.add_argument(
            "--ignore-extractor-not-found",
            dest="ignore_extractor_not_found",
            action="store_true",
            default=get_bool_env("MIGRATION_LINTER_IGNORE_EXTRACTOR_NOT_FOUND", False),
            help="""
                Don't fail the whole linter if extraction went fine,
                but info about particular migration couldn't be found
            """,
        )

    def handle(self, loader_type, squawk_config_path, squawk_pg_version, **options):
        logger.info("Start analysis..")

        loader = SourceLoader.get(loader_type)(**options)
        extractor = Extractor.get("django_management")(**options)
        analyzer = Analyzer(
            loader=loader,
            extractor=extractor,
            linters=[
                CompatibilityLinter(),
                SquawkLinter(
                    config_path=squawk_config_path,
                    pg_version=squawk_pg_version,
                ),
            ],
        )
        analyzer.analyze()



import os
import re
import subprocess
from io import StringIO
from functools import lru_cache

from migration_lint import logger
from migration_lint.extractor.base import BaseExtractor


class AlembicExtractor(BaseExtractor):
    """Migrations extractor for Alembic migrations."""

    NAME = "alembic"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.command = kwargs.get("alembic_command") or "make sqlmigrate"
        self.migration_path = os.environ.get(
            "MIGRATION_LINT_ALEMBIC_MIGRATIONS_PATH", "/migrations/versions/"
        )

    def is_migration(self, path: str) -> bool:
        """Check if the specified file is a migration."""

        return (
            self.migration_path in path
            and path.endswith(".py")
            and "__init__.py" not in path
        )

    def is_allowed_with_backward_incompatible_migration(self, path: str) -> bool:
        """Check if the specified file changes are allowed with
        backward-incompatible migrations.
        """

        allowed_patterns = [
            r".*/tables\.py",
            r".*/constants\.py",
            r".*/enums\.py",
            rf".*{self.migration_path}.*\.py",
            r"^(?!.*\.py).*$",
        ]
        for pattern in allowed_patterns:
            if re.match(pattern, path):
                return True

        return False

    def extract_sql(self, migration_path: str) -> str:
        """Extract raw SQL from the migration file."""

        file_name = migration_path.split("/")[-1]
        parts = file_name.split("_")
        version = parts[1]

        logger.info(f"Extracting sql for migration: version={version}")
        logger.info(self.command)

        migrations_sql = self._get_migrations_sql()

        try:
            return migrations_sql[version]
        except KeyError:
            error_msg = (
                f"Couldn't find info about migration with version={version} "
                f"in alembic offline mode output"
            )
            if self.ignore_extractor_not_found:
                logger.error(error_msg)
                return ""
            raise RuntimeError(error_msg)

    @lru_cache(maxsize=1)
    def _get_migrations_sql(self):
        """Get raw SQL for all migrations."""

        try:
            lines = (
                subprocess.check_output(self.command.split(" "))
                .decode("utf-8")
                .split("\n")
            )
        except subprocess.CalledProcessError:
            logger.error("Failed to extract SQL for migrations")
            if self.ignore_extractor_fail:
                return {}
            raise

        migrations_sql = {}

        current_migration = None
        current_migration_sql = StringIO()

        for line in lines:
            m = re.match(r"-- Running upgrade \w* -> (\w*)", line)
            if m is not None:
                if current_migration is not None:
                    migrations_sql[current_migration] = (
                        current_migration_sql.getvalue().strip("\n")
                    )

                current_migration = m.group(1)
                current_migration_sql = StringIO()

            elif "UPDATE alembic_version" in line:
                continue
            elif line.startswith("/"):
                continue
            else:
                current_migration_sql.write(f"{line}\n")

        else:
            if current_migration is not None:
                migrations_sql[current_migration] = (
                    current_migration_sql.getvalue().strip("\n")
                )

        return migrations_sql



from __future__ import annotations

import abc
from typing import Any, Sequence, cast, Dict

from migration_lint.source_loader.model import SourceDiff


class SourceLoader(type):
    """Metaclass for source loaders.

    This metaclass register all its instances in the registry.
    """

    source_loaders: Dict[str, SourceLoader] = {}

    def __new__(
        mcls,
        name: str,
        bases: tuple[SourceLoader, ...],
        classdict: Dict[str, Any],
    ) -> SourceLoader:
        cls = cast(SourceLoader, type.__new__(mcls, name, bases, classdict))

        if len(bases) > 0:
            # Not the base class.
            if not hasattr(cls, "NAME"):
                raise NotImplementedError(
                    f"source loader {cls.__name__} doesn't provide name",
                )

            mcls.source_loaders[cls.NAME] = cls

        return cls

    @classmethod
    def names(mcls) -> Sequence[str]:
        """Get the names of all registered source loaders."""

        return list(mcls.source_loaders.keys())

    @classmethod
    def get(mcls, name: str) -> SourceLoader:
        """Get a registered source loader by its name."""

        return mcls.source_loaders[name]


class BaseSourceLoader(metaclass=SourceLoader):
    """Base class for changes source files loader."""

    def __init__(self, only_new_files: bool, **kwargs: Any) -> None:
        self.only_new_files = only_new_files

    @abc.abstractmethod
    def get_changed_files(self) -> Sequence[SourceDiff]:
        """Return a list of changed files."""

        raise NotImplementedError()



import logging

logger = logging.getLogger("default")
logger.setLevel(level=logging.DEBUG)
fh = logging.StreamHandler()
fh_formatter = logging.Formatter("%(message)s")
fh.setFormatter(fh_formatter)
logger.addHandler(fh)



import re
import subprocess

from migration_lint import logger
from migration_lint.extractor.base import BaseExtractor


class DjangoExtractor(BaseExtractor):
    """Migrations extractor for Django migrations."""

    NAME = "django"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # self.command = "python manage.py sqlmigrate {app} {migration_name}"
        self.command = "make sqlmigrate app={app} migration={migration_name}"
        self.skip_lines = 2

    def is_migration(self, path: str) -> bool:
        """Check if the specified file is a migration."""

        return (
            "/migrations/" in path
            and path.endswith(".py")
            and "__init__.py" not in path
        )

    def is_allowed_with_backward_incompatible_migration(self, path: str) -> bool:
        """Check if the specified file changes are allowed with
        backward-incompatible migrations.
        """

        allowed_patterns = [
            r".*/models\.py",
            r".*/constants\.py",
            r".*/enums\.py",
            r".*/migrations/.*\.py",
            r"^(?!.*\.py).*$",
        ]
        for pattern in allowed_patterns:
            if re.match(pattern, path):
                return True

        return False

    def extract_sql(self, migration_path: str) -> str:
        """Extract raw SQL from the migration file."""

        parts = migration_path.split("/")
        file_name = parts[-1]
        app = parts[-3]
        migration_name = file_name.replace(".py", "")

        logger.info(
            f"Extracting sql for migration: app={app}, migration_name={migration_name}"
        )

        try:
            output = subprocess.check_output(
                self.command.format(app=app, migration_name=migration_name).split(" ")
            ).decode("utf-8")
        except subprocess.CalledProcessError:
            logger.error(
                f"Failed to extract SQL for migration app={app}, migration_name={migration_name}"
            )
            if self.ignore_extractor_fail:
                return ""
            raise
        return "\n".join(output.split("\n")[self.skip_lines :])



from __future__ import annotations

import abc
from typing import Any, Sequence, cast, Dict

from migration_lint.source_loader.model import SourceDiff
from migration_lint.extractor.model import (
    ExtendedSourceDiff,
    Migration,
    MigrationsMetadata,
)


class Extractor(type):
    """Metaclass for migrations extractors.

    This metaclass registers all its instances in the registry.
    """

    extractors: Dict[str, Extractor] = {}

    def __new__(
        mcls,
        name: str,
        bases: tuple[Extractor, ...],
        classdict: dict[str, Any],
    ) -> Extractor:
        cls = cast(Extractor, type.__new__(mcls, name, bases, classdict))

        if len(bases) > 0:
            # Not the base class.
            if not hasattr(cls, "NAME"):
                raise NotImplementedError(
                    f"extractor {cls.__name__} doesn't provie name",
                )

            mcls.extractors[cls.NAME] = cls

        return cls

    @classmethod
    def names(mcls) -> Sequence[str]:
        """Get the names of all registered extractors."""

        return list(mcls.extractors.keys())

    @classmethod
    def get(mcls, name: str) -> Extractor:
        """Get a registered extractor by its name."""

        return mcls.extractors[name]


class BaseExtractor(metaclass=Extractor):
    """Base class for migrations extractor."""

    def __init__(self, **kwargs) -> None:
        self.ignore_extractor_fail = kwargs.get("ignore_extractor_fail")
        self.ignore_extractor_not_found = kwargs.get("ignore_extractor_not_found")

    def create_metadata(
        self,
        changed_files: Sequence[SourceDiff],
    ) -> MigrationsMetadata:
        """Create migrations metadata by the list of changed files."""

        metadata = MigrationsMetadata()
        for changed_file in changed_files:
            path = changed_file.path

            metadata.changed_files.append(
                ExtendedSourceDiff.of_source_diff(
                    changed_file,
                    self.is_allowed_with_backward_incompatible_migration(path),
                ),
            )

            if self.is_migration(path):
                metadata.migrations.append(
                    Migration(path=path, raw_sql=self.extract_sql(path)),
                )

        return metadata

    @abc.abstractmethod
    def is_migration(self, path: str) -> bool:
        """Check if the specified file is a migration."""

        raise NotImplementedError()

    @abc.abstractmethod
    def is_allowed_with_backward_incompatible_migration(self, path: str) -> bool:
        """Check if the specified file changes are allowed with
        backward-incompatible migrations.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def extract_sql(self, migration_path: str) -> str:
        """Extract raw SQL from the migration file."""

        raise NotImplementedError()



from __future__ import annotations

import os.path
import re
import subprocess
import sys
from io import StringIO
from collections import defaultdict
from typing import List, Dict, Optional

import migration_lint
from migration_lint.analyzer.base import BaseLinter
from migration_lint.extractor.model import ExtendedSourceDiff


class SquawkLinter(BaseLinter):
    """Squawk linter integration."""

    ignored_rules = [
        "ban-drop-column",  # Backward-incompatible, checked by compatibility analyzer.
        "ban-drop-table",  # Backward-incompatible, checked by compatibility analyzer.
        "adding-not-nullable-field",  # Backward-incompatible, checked by compatibility analyzer.
        "prefer-big-int",  # Deprecated.
        "prefer-identity",
        "prefer-timestamptz",
        "ban-drop-not-null",  # Dropping a NOT NULL constraint is safe but may break existing clients.
        "prefer-robust-stmts",  # TODO: Add transactions tracking.
        "transaction-nesting",  # TODO: Add transactions tracking.
    ]

    def __init__(
        self, config_path: Optional[str] = None, pg_version: Optional[str] = None
    ) -> None:
        bin_dir = os.path.join(migration_lint.__path__[0], "bin")
        self.config_path = config_path
        self.pg_version = pg_version
        if sys.platform == "linux":
            self.squawk = os.path.join(bin_dir, "squawk-linux-x86")
        elif sys.platform == "darwin":
            self.squawk = os.path.join(bin_dir, "squawk-darwin-arm64")
        else:
            raise RuntimeError(f"unsupported platform: {sys.platform}")

    def squawk_command(self, migration_sql: str) -> str:
        """Get squawk command."""
        exclude = f"--exclude={','.join(self.ignored_rules)}"
        config = f"--config={self.config_path}" if self.config_path else ""
        pg_version = f"--pg-version={self.pg_version}" if self.pg_version else ""

        return " ".join([self.squawk, exclude, config, pg_version]).strip()

    def lint(
        self,
        migration_sql: str,
        changed_files: List[ExtendedSourceDiff],
    ) -> List[str]:
        """Perform SQL migration linting."""

        output = subprocess.run(
            self.squawk_command(migration_sql),
            input=migration_sql.encode(),
            shell=True,
            stdout=subprocess.PIPE,
        ).stdout.decode("utf-8")

        # Reformat the output for brevity.

        statements: Dict[str, str] = {}
        error_msgs: Dict[str, List[str]] = defaultdict(list)

        key = None
        statement = None
        error_msg = None
        for line in output.splitlines():
            stmt_m = re.match(r"\s+\d+\s+\|\s(.*)", line)

            if line.startswith("stdin:"):
                if key is not None:
                    statements[key] = statement.getvalue()
                    error_msgs[key].append(error_msg.getvalue())

                key = line[: line.index(" ")]

                statement = None
                error_msg = StringIO()
                error_msg.write(f"  - squawk {line[line.index(' ') + 1:]}\n\n")

            elif stmt_m is not None:
                if statement is None:
                    statement = StringIO()
                    statement.write(f"- {stmt_m.group(1)}\n")
                else:
                    statement.write(f"  {stmt_m.group(1)}\n")

            elif line.startswith("  "):
                if error_msg is not None:
                    error_msg.write(f"  {line}\n")
        else:
            if key is not None:
                assert statement is not None
                assert error_msg is not None

                statements[key] = statement.getvalue()
                error_msgs[key].append(error_msg.getvalue())

        # Generate the errors.

        errors: List[str] = []

        error = StringIO()
        for key in statements:
            error.write(f"{statements[key]}\n")
            for msg in error_msgs[key]:
                error.write(msg)
            errors.append(error.getvalue())

            error = StringIO()

        if errors:
            errors.append(
                "squawk: find detailed examples and solutions for each rule at "
                "https://squawkhq.com/docs/rules"
            )

        return errors



import os
from typing import Sequence

from migration_lint import logger
from migration_lint.source_loader.base import BaseSourceLoader
from migration_lint.source_loader.model import SourceDiff


class LocalLoader(BaseSourceLoader):
    """A loader to obtain files changed for local stashed files."""

    NAME = "local_git"

    def get_changed_files(self) -> Sequence[SourceDiff]:
        """Return a list of changed files."""

        from git import Repo

        logger.info("### Getting changed files for local stashed files")

        repo = Repo(os.getcwd(), search_parent_directories=True)
        diffs = repo.head.commit.diff(None)
        filtered_diffs = [
            d
            for d in diffs
            if not d.deleted_file
            and (not self.only_new_files or self.only_new_files and d.new_file)
        ]

        logger.info("Files changed: ")
        logger.info("\n".join([f"- {d.a_path}" for d in filtered_diffs]))

        return [
            SourceDiff(old_path=diff.a_path, path=diff.b_path)
            for diff in filtered_diffs
        ]



from migration_lint.extractor.base import BaseExtractor


class RawSqlExtractor(BaseExtractor):
    """Migrations extractor for SQL files migrations."""

    NAME = "raw_sql"

    def is_migration(self, path: str) -> bool:
        """Check if the specified file is a migration."""

        return path.endswith(".sql")

    def is_allowed_with_backward_incompatible_migration(self, path: str) -> bool:
        """Check if the specified file changes are allowed with
        backward-incompatible migrations.
        """

        return self.is_migration(path)

    def extract_sql(self, migration_path: str) -> str:
        """Extract raw SQL from the migration file."""

        try:
            with open(migration_path, "r") as f:
                return f.read()
        except:
            if self.ignore_extractor_fail:
                return ""
            raise



from migration_lint.source_loader.base import SourceLoader
from migration_lint.source_loader.local import LocalLoader
from migration_lint.source_loader.gitlab import GitlabBranchLoader, GitlabMRLoader

__all__ = (
    "SourceLoader",
    "LocalLoader",
    "GitlabBranchLoader",
    "GitlabMRLoader",
)



"""
Migration linter for Django app.

This app has `django` management command
that efficiently handle MRs with many migrations.
"""

default_app_config = "migration_lint.django.apps.MigrationLintDjangoConfig"



from __future__ import annotations

from io import StringIO
from typing import List

from migration_lint import logger
from migration_lint.analyzer.base import BaseLinter
from migration_lint.extractor.model import ExtendedSourceDiff
from migration_lint.sql.constants import StatementType
from migration_lint.sql.parser import classify_migration
from migration_lint.util.colors import blue


DOCS_URL = "https://pandadoc.github.io/migration-lint/classification/"


class CompatibilityLinter(BaseLinter):
    f"""
    Custom linter that checks backward compatibility
    based on migrations classification.

    See {DOCS_URL} for details.
    """

    def lint(
        self,
        migration_sql: str,
        changed_files: List[ExtendedSourceDiff],
        report_restricted: bool = False,
    ) -> List[str]:
        """Perform SQL migration linting."""

        errors = []

        classification_result = classify_migration(migration_sql)

        statement_types = set()
        logger.info(blue("Migration contains statements:\n"))
        for statement_sql, statement_type in classification_result:
            statement_types.add(statement_type)
            logger.info(f"- {statement_type.colorized}: {statement_sql}")

            if statement_type == StatementType.UNSUPPORTED:
                errors.append(f"- Statement can't be identified: {statement_sql}")

            if statement_type == StatementType.RESTRICTED and report_restricted:
                errors.append(
                    (
                        f"- Statement is restricted to use: {statement_sql}."
                        f"\n\tCheck the doc to do this correctly: {DOCS_URL}.\n"
                    )
                )

        if StatementType.RESTRICTED in statement_types and not report_restricted:
            errors.append(
                (
                    "- There are restricted statements in migration"
                    "\n\tCheck squawk output below for details"
                    f"\n\tAlso check the doc to fix it: {DOCS_URL}\n"
                )
            )

        if StatementType.BACKWARD_INCOMPATIBLE in statement_types:
            not_allowed_files = [
                file.path
                for file in changed_files
                if not file.allowed_with_backward_incompatible
            ]
            if not_allowed_files:
                error = StringIO()
                error.write(
                    (
                        "- You have backward incompatible operations, "
                        "which is not allowed with changes in following files:"
                    )
                )
                for file_name in not_allowed_files:
                    error.write(f"\n\t- {file_name}")
                error.write(
                    "\n\n\tPlease, separate changes in different merge requests.\n"
                )

                errors.append(error.getvalue())

        if StatementType.DATA_MIGRATION in statement_types and (
            StatementType.BACKWARD_COMPATIBLE in statement_types
            or StatementType.BACKWARD_INCOMPATIBLE in statement_types
        ):
            statement_sql = [
                r[0]
                for r in classification_result
                if r[1] == StatementType.DATA_MIGRATION
            ][0]
            errors.append(
                (
                    f"- Seems like you have data migration along with schema migration: {statement_sql}"
                    "\n\n\tPlease, separate changes in different merge requests.\n"
                )
            )

        return errors



import subprocess


SQUAWK_VERSION = "1.1.2"


if __name__ == "__main__":
    subprocess.run(["mkdir", "-p", "migration_lint/bin"])
    subprocess.run(
        [
            "curl",
            "-L",
            f"https://github.com/sbdchd/squawk/releases/download/v{SQUAWK_VERSION}/squawk-darwin-arm64",
            "-o",
            "migration_lint/bin/squawk-darwin-arm64",
        ]
    )
    subprocess.run(
        [
            "curl",
            "-L",
            f"https://github.com/sbdchd/squawk/releases/download/v{SQUAWK_VERSION}/squawk-linux-x64",
            "-o",
            "migration_lint/bin/squawk-linux-x86",
        ]
    )
    subprocess.run(["chmod", "+x", "migration_lint/bin/squawk-darwin-arm64"])
    subprocess.run(["chmod", "+x", "migration_lint/bin/squawk-linux-x86"])
